/*==================================================================================
| code\sp2cpu.v: sp2 processor core                                                |
|                                                                                  |
|           Copyright 2000-2007 by John Rible <sp2@sandpipers.com>                 |
|  Permission is granted for individual non-commercial use, provided this notice   |
|  is included. Contact the author concerning any other use.                       |
|                                                                                  |
|  My Verilog name conventions:  - nets and ports have lower-case names            |
|     80 lines per page          - registers and flipflops have UPPERCASE NAMES    |
|     84 characters per line     - functions and tasks have Capitalized Names      |
====================================================================================

   The sp2 cpu follows the structure of Chuck Moore's series of '-21'
   processors and the related 16-bit FPGA designs produced by Dr. Ting. For more
   information, visit Jeff Fox's web site at <http://www.Ultratechnology.com>.

   Features common to all these processors include:

   o  5-bit opcodes packed into a larger datapath word size,
   o  separate on-chip operand and return stacks outside addressable memory space,
   o  single-cycle opcode execution, with no cache or pipeline,
   o  jump/call opcodes are variable-length,
   o  immediate address bits replace base register bits (faster than adding),
   o  two conditional jumps: top-of-stack(TOS) = zero and CARRY = zero.
   o  all i/o and co-processor access is memory-mapped and word addressed

   Features added to or changed on the sp2 cpu include:

   o  hardware 'bit-threaded code' intrepreter included
   o  use of 1-bit return modifier to add conditional call and repeat opcodes.
   o  making the shift and add opcodes double length, modifying both A and T
   o  data load/store with the operand stack (like the NOVIX 4000 & Harris RTX2000)
   o  opcodes are closer to "traditional Forth" than most "machine Forth" versions
------------------------------------------------------------------------------------
20001116 jwr created this file.
20020823 jwr separated decode, incrementer, cleaned up for Fall class
20070716 jwr began BEI edits: changed paths and parameters
20070718 jwr merged opcode and alu modules into cpu
20070725 jwr another round of opcode name edits; move relt to rep; optional nxtl
20070730 jwr cpu coding complete--all except jmp-group has been exercised
20070730 jwr all code and List opcodes have been exercised & appear to work
20070806 jwr code tweaks--looking for speedups, included incrementer
20070811 jwr moved conditionals to datapath
20070817 jwr changed to OAR: names, alu ops, XTs
20070822 jwr changed opcodes to forth names, DS,O -> DS,D
==================================================================================*/





module sp2( clock, reset, hold,  data, read, next,  cda, cdo, cdi, done);

   input         clock;   wire        clock;
   input         reset;   wire        reset;
   input         hold;    wire        hold;   // high to disable ma_bus increment
                                              // mem access: none fetch write read
   output        data;                        //       data:  0     0     1     1
   output        read;                        //       read:  0     1     0     1
   output        next;                        // high is List

   output [15:0] cda;     wire [15:0] cda;    // cpu data address
   output [15:0] cdo;     wire [15:0] cdo;    // cpu data out
   input  [15:0] cdi;     wire [15:0] cdi;    // cpu data in
   input         done;    wire        done;   // set when memory access is complete

   reg    [15:0] I_REG, P_REG, R_REG, A_REG, D_REG ;
   reg     [3:0] nextSLOT, SLOT;
   reg     [4:0] nextOPCODE, OPCODE;
   reg           LIST, RNZERO, ANZERO, ANEG;

`define tests    jmp, jsr, rep, unconditional, test_rn, test_ap, test_az
`define early    sel_ra, sel_ma, read, data, dec_ra
`define middle   push_rs, pop_rs, push_ds, pop_ds
`define selects  sel_ps, sel_ri, sel_dn, sel_ai, sel_dd, sel_di, inc_slot
`define enables  new_D, new_A, new_R, new_P, new_I, next
`define controls `tests, `early, `middle, `selects, `enables

   reg `controls;


// LIST state           :-tests--:--early------:-mid-:--------selects------:enables
//                      :grp cond ra ma rd dec :RS DS: ps ri dn ai dd di  s DARPI L

`define decode_exit 29'b 000_0000__1__0_10__x___01_00___0__1__x__x__x__x__0_00111_1
`define decode_nest 29'b 000_0000__x__1_10__x___10_00___1__1__x__x__x__x__0_00111_1
`define decode_call 29'b 000_0000__x__1_10__x___10_00___1__1__x__x__x__x__0_00111_0

// CODE state           :-tests--:--early------:-mid-:--------selects------:enables
// jump group opcodes   :grp cond ra ma rd dec :RS DS: ps ri dn ai dd di  s DARPI L

`define decode_jmp  29'b 100_1000__x__1_10__x___00_00___x__x__x__x__x__x__0_00011_0
`define decode_jmpd 29'b 100_0100__1__1_10__1___01_00___0__1__0__x__x__x__0_00111_0
   // when jmpd taken   :----[ma]--1--------1----0---------0-----------------------
`define decode_jmpp 29'b 100_0010__x__1_10__x___00_00___x__x__x__x__x__x__0_00011_0
`define decode_jmpz 29'b 100_0001__x__1_10__x___00_00___x__x__x__x__x__x__0_00011_0
   // when jmp? taken   :----[ma]--------------------------------------------------

`define decode_jsr  29'b 110_1000__x__1_10__x___00_00___x__1__x__x__x__x__0_00011_0
`define decode_jsrn 29'b 110_0100__x__1_10__x___00_00___x__1__x__x__x__x__0_00011_0
`define decode_jsrp 29'b 110_0010__x__1_10__x___00_00___x__1__x__x__x__x__0_00011_0
`define decode_jsrz 29'b 110_0001__x__1_10__x___00_00___x__1__x__x__x__x__0_00011_0
   // when jsr? taken   :----[ma]---------------1-------1---------------------1----

`define decode_repd 29'b 001_0100__1__x_00__1___01_00___0__1__0__x__x__x__1_00100_0
   // when repd taken:   ----------1--------1----0---------0--------------0--------
`define decode_repp 29'b 001_0010__x__x_00__x___00_00___x__x__x__x__x__x__1_00000_0
`define decode_repz 29'b 001_0001__x__x_00__x___00_00___x__x__x__x__x__x__1_00000_0
   // when rep? taken   :-------------------------------------------------0--------

`define decode_nxt  29'b 000_0000__x__1_10__x___00_00___x__x__x__x__x__x__0_00011_0
`define decode_into 29'b 000_0000__x__1_10__x___00_00___x__x__x__x__x__x__0_00011_1

`define decode_ret  29'b 000_0000__1__0_10__x___01_00___0__1__x__x__x__x__0_00111_0
`define decode_back 29'b 000_0000__1__0_10__x___01_00___0__1__x__x__x__x__0_00111_1

// move group opcodes   :grp cond ra ma rd dec :RS DS: ps ri dn ai dd di  s DARPI L

`define decode_get  29'b 000_0000__x__x_00__x___00_10___x__x__0__1__x__1__1_11000_0
`define decode_pop  29'b 000_0000__1__x_00__0___01_10___0__1__0__1__x__1__1_11100_0
`define decode_drop 29'b 000_0000__x__x_00__x___00_01___x__x__x__0__0__0__1_11000_0
`define decode_push 29'b 000_0000__0__x_00__0___10_01___x__0__0__0__0__0__1_11100_0

// memory group opcodes :grp cond ra ma rd dec :RS DS: ps ri dn ai dd di  s DARPI L

`define decode_lit  29'b 000_0000__x__1_11__x___00_10___x__x__x__x__1__0__1_10010_0
`define decode_ldri 29'b 000_0000__1__0_11__x___00_10___x__0__1__x__1__0__1_10100_0
`define decode_lda  29'b 000_0000__0__0_11__x___00_10___x__x__x__x__1__0__1_10000_0
`define decode_ldai 29'b 000_0000__0__0_11__x___00_10___x__x__1__1__1__0__1_11000_0

`define decode_str  29'b 000_0000__1__0_01__x___00_01___x__x__x__x__0__0__1_10000_0
`define decode_stri 29'b 000_0000__1__0_01__x___00_01___x__0__1__x__0__0__1_10100_0
`define decode_sta  29'b 000_0000__0__0_01__x___00_01___x__x__x__x__0__0__1_10000_0
`define decode_stai 29'b 000_0000__0__0_01__x___00_01___x__x__1__1__0__0__1_11000_0

// alu group opcodes    :grp cond ra ma rd dec :RS DS: ps ri dn ai dd di  s DARPI L

`define decode_swap 29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0
`define decode_nop  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0
`define decode_eqz  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0
`define decode_inv  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0

`define decode_add  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0
`define decode_lsl  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0
`define decode_asr  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0
`define decode_lsr  29'b 000_0000__x__x_00__x___00_00___x__x__x__0__x__1__1_11000_0

`define decode_over 29'b 000_0000__x__x_00__x___00_10___x__x__x__0__x__1__1_11000_0
`define decode_dup  29'b 000_0000__x__x_00__x___00_10___x__x__x__0__x__1__1_11000_0
`define decode_zero 29'b 000_0000__x__x_00__x___00_10___x__x__x__0__x__1__1_11000_0
`define decode_one  29'b 000_0000__x__x_00__x___00_10___x__x__x__0__x__1__1_11000_0

`define decode_nip  29'b 000_0000__x__x_00__x___00_01___x__x__x__0__0__0__1_11000_0
`define decode_or   29'b 000_0000__x__x_00__x___00_01___x__x__x__0__0__0__1_11000_0
`define decode_xor  29'b 000_0000__x__x_00__x___00_01___x__x__x__0__0__0__1_11000_0
`define decode_and  29'b 000_0000__x__x_00__x___00_01___x__x__x__0__0__0__1_11000_0

`define undefined   29'b xxx_xxxx__x__x_xx__x___xx_xx___x__x__x__x__x__x__x_xxxxx_x
//                      :grp cond ra ma rd dec :RS DS: ps ri dn ai dd di  s DARPI L
//                      :-tests--:--early------:-mid-:--------selects------:enables

// decode LIST tokens and CODE opcodes ---------------------------------------------

   always @* begin: decode
     casex ({LIST,SLOT[3]})

       2'b0_0: casex (OPCODE)  // Code State slot 0-2

         5'b1xxxx: case (OPCODE[3:0]) // alu group

           4'b0000:  {`controls} = `decode_swap; // $10
           4'b0001:  {`controls} = `decode_nop ; // $11
           4'b0010:  {`controls} = `decode_eqz ; // $12
           4'b0011:  {`controls} = `decode_inv ; // $13

           4'b0100:  {`controls} = `decode_add ; // $14
           4'b0101:  {`controls} = `decode_lsl ; // $15
           4'b0110:  {`controls} = `decode_asr ; // $16
           4'b0111:  {`controls} = `decode_lsr ; // $17

           4'b1000:  {`controls} = `decode_over; // $18
           4'b1001:  {`controls} = `decode_dup ; // $19
           4'b1010:  {`controls} = `decode_zero; // $1a
           4'b1011:  {`controls} = `decode_one ; // $1b

           4'b1100:  {`controls} = `decode_nip ; // $1c
           4'b1101:  {`controls} = `decode_or  ; // $1d
           4'b1110:  {`controls} = `decode_xor ; // $1e
           4'b1111:  {`controls} = `decode_and ; // $1f

           default:  {`controls} = `undefined; endcase

         5'b01xxx: case (OPCODE[2:0]) // memory group

           3'b000:   {`controls} = `decode_lit ; // $08
           3'b001:   {`controls} = `decode_ldri; // $09
           3'b010:   {`controls} = `decode_lda ; // $0a
           3'b011:   {`controls} = `decode_ldai; // $0b

           3'b100:   {`controls} = `decode_str ; // $0c
           3'b101:   {`controls} = `decode_stri; // $0d
           3'b110:   {`controls} = `decode_sta ; // $0e
           3'b111:   {`controls} = `decode_stai; // $0f

           default:  {`controls} = `undefined; endcase

         5'b001xx: case (OPCODE[1:0]) // move group

           2'b00:    {`controls} = `decode_get ; // $04
           2'b01:    {`controls} = `decode_pop ; // $05
           2'b10:    {`controls} = `decode_drop; // $06
           2'b11:    {`controls} = `decode_push; // $07

           default:  {`controls} = `undefined; endcase

         5'b000xx: casex ({SLOT[2],I_REG[15],OPCODE[1:0]}) // control group

           4'b00_00: {`controls} = `decode_jmp ; // $00
           4'b00_01: {`controls} = `decode_jmpd; // $01
           4'b00_10: {`controls} = `decode_jmpp; // $02
           4'b00_11: {`controls} = `decode_jmpz; // $03

           4'b01_00: {`controls} = `decode_jsr ; // $00
           4'b01_01: {`controls} = `decode_jsrn; // $01
           4'b01_10: {`controls} = `decode_jsrp; // $02
           4'b01_11: {`controls} = `decode_jsrz; // $03

           4'b10_00: {`controls} = `decode_into; // $00 enter inline LIST
           4'b10_01: {`controls} = `decode_repd; // $01
           4'b10_10: {`controls} = `decode_repp; // $02
           4'b10_11: {`controls} = `decode_repz; // $03

           4'b11_00: {`controls} = `decode_back; // $00 return to previous LIST
           4'b11_01: {`controls} = `decode_repd; // $01
           4'b11_10: {`controls} = `decode_repp; // $02
           4'b11_11: {`controls} = `decode_repz; // $03

           default:  {`controls} = `undefined; endcase

         default:    {`controls} = `undefined; endcase // Code State slot 0-2

       2'b0_1: case (I_REG[15])  // Code State slot 3

         1'b0:       {`controls} = `decode_nxt ; // fetch next CODE instruction
         1'b1:       {`controls} = `decode_ret ; // return to previous CODE

         default:    {`controls} = `undefined; endcase // Code State slot 3

       2'b1_x: casex ({|I_REG[15:1],I_REG[0]})  // List State

         2'b0_0:     {`controls} = `decode_exit; // return to previous LIST
         2'b1_0:     {`controls} = `decode_call; // call to CODE routine
         2'bx_1:     {`controls} = `decode_nest; // call to LIST routine

         default:    {`controls} = `undefined; endcase // List State

       default:      {`controls} = `undefined;
     endcase // state,slot3
   end // decode









// list state and opcode slot handling ---------------------------------------------

   reg           dec_R, taken, jumping, calling, repeating;
   reg           sel_fa, sel_95, sel_40, enable;


   always @* begin: decide

     dec_R = RNZERO & test_rn;
     taken = unconditional | dec_R | ~ANZERO & test_az | ~ANEG & test_ap;

     jumping = jmp & taken;           // both jumps and calls
     calling = jsr & taken;
     repeating = rep & taken;

     sel_fa = LIST;
     sel_95 = LIST | jumping & SLOT[0];
     sel_40 = LIST | jumping & SLOT[0] | jumping & SLOT[1];

     nextSLOT = repeating | ~inc_slot? 4'b 0001: // to slot 0 on new instruction
                                        SLOT<<1; // otherwise increment slot
     casex (nextSLOT)
       4'bxxx1: nextOPCODE = repeating? I_REG[14:10]: cdi[14:10];
       4'bxx10: nextOPCODE = I_REG[9:5];
       4'bx100: nextOPCODE = I_REG[4:0];
       default: nextOPCODE = 5'bxxxxx;
     endcase

     enable = ~data & ~read | done;

   end // decide


   always @(posedge clock or posedge reset) begin: state

      if (reset) begin
         LIST   <=  1'b     0;   // start in CODE state
         SLOT   <=  4'b  1000;   // start in slot 3
         OPCODE <=  5'b 00000;
      end
      else if (enable) begin
         LIST   <= next;
         SLOT   <= nextSLOT;
         OPCODE <= nextOPCODE;
      end
   end // state







/* alu -----------------------------------------------------------------------------

  The alu functions are limited only by Chuck Moore's desire to have the minimum
  hardware needed to perform useful work. By generating all the possible outputs and
  selecting the proper one at the end, he does the decode in parallel with the add.

  In this implementation, all alu functions complete in a single cycle, involving
  only the accumulator and operand stack top register.

----------------------------------------------------------------------------------*/

   reg    [15:0] zd_bus, za_bus;

   always @* begin: alu

     reg [15:0] eqz_bus, or_bus, xor_bus, and_bus;
     reg [16:0] sum_bus;

     eqz_bus = ANZERO? 16'h0000: 16'hffff;

     sum_bus = A_REG + D_REG;   // the 17-bit sum includes the carry

      or_bus = A_REG | D_REG;
     xor_bus = A_REG ^ D_REG;
     and_bus = A_REG & D_REG;

     case (OPCODE[4]? OPCODE[3:0]: 4'b0000) // only ALU opcodes

       4'b0000: {zd_bus,za_bus} = {A_REG,  D_REG};               // exch  single
       4'b0001: {zd_bus,za_bus} = {D_REG,  A_REG};               // nop
       4'b0010: {zd_bus,za_bus} = {D_REG,eqz_bus};               // eqz
       4'b0011: {zd_bus,za_bus} = {D_REG, ~A_REG};               // inv

       4'b0100: {zd_bus,za_bus} = {15'b0,sum_bus};               // add   double
       4'b0101: {zd_bus,za_bus} = {D_REG[14:0],    A_REG, 1'b0}; // shl
       4'b0110: {zd_bus,za_bus} = {D_REG[15],D_REG,A_REG[15:1]}; // sar
       4'b0111: {zd_bus,za_bus} = {     1'b0,D_REG,A_REG[15:1]}; // shr

       4'b1000: {zd_bus,za_bus} = {A_REG,  D_REG};               // geto  push
       4'b1001: {zd_bus,za_bus} = {A_REG,  A_REG};               // geta
       4'b1010: {zd_bus,za_bus} = {A_REG,  16'd0};               // zero
       4'b1011: {zd_bus,za_bus} = {A_REG,  16'd1};               // one

       4'b1100: {zd_bus,za_bus} = {A_REG,  A_REG};               // delo  pop
       4'b1101: {zd_bus,za_bus} = {A_REG, or_bus};               // or
       4'b1110: {zd_bus,za_bus} = {A_REG,xor_bus};               // xor
       4'b1111: {zd_bus,za_bus} = {A_REG,and_bus};               // and

       default: {zd_bus,za_bus} = {A_REG,  D_REG};
     endcase

   end // alu






// instantiate stacks --------------------------------------------------------------

   wire   [15:0] rs_bus,  ds_bus;           // stack access: idle push pop error
                                            //    *s_pop   :  0    0    1    1
                                            //    *s_push  :  0    1    0    1

   stack rs( .reset(reset), .clock(clock), .enable(enable),
             .push(push_rs|calling), .push_bus(R_REG),
             .pop(pop_rs & ~dec_R), .pop_bus(rs_bus) );

   stack ds( .reset(reset), .clock(clock), .enable(enable),
             .push(push_ds), .push_bus(D_REG),
             .pop(pop_ds), .pop_bus(ds_bus) );









// datapath busses and registers ---------------------------------------------------

   reg    [15:0] ra_bus, ip_bus, pi_bus, ii_bus, ma_bus;
   reg    [15:0] dr_bus, dn_bus, ps_bus, dd_bus, ri_bus, ai_bus, di_bus;

   always @* begin: sooner

     ip_bus = {(sel_fa?  I_REG[15:10]: P_REG[15:10]),
               (sel_95?  I_REG[09:05]: P_REG[09:05]),
               (sel_40?  I_REG[04:00]: P_REG[04:00])};

     ra_bus = sel_ra?  R_REG:        A_REG;
     ma_bus = sel_ma? ip_bus:       ra_bus;

     ii_bus = cdi;

   end // sooner


   always @* begin: later

     reg [16:0] dec_bus, inc_bus;

     inc_bus = ma_bus + 16'd1;
     dec_bus =  R_REG - 16'd1;

     pi_bus = hold?    ma_bus:      inc_bus[15:0];  // passes ma or ma+1
     dr_bus = dec_ra? dec_bus[15:0]: ra_bus;        // passes ra-1 or ra
     dn_bus = sel_dn?  pi_bus:       dr_bus;

     ps_bus = sel_ps | calling?   P_REG:  rs_bus;   // jjj
     ri_bus = sel_ri & ~dec_R?   ps_bus:  dn_bus;

     ai_bus = sel_ai?  dn_bus:  za_bus;

     dd_bus = sel_dd?  ii_bus:  ds_bus;
     di_bus = sel_di?  zd_bus:  dd_bus;

   end // later

   always @(posedge clock or posedge reset) begin: regs

     if (reset) begin

       D_REG  <= 16'h 0000;
       A_REG  <= 16'h 0000;
       ANZERO <=  1'b    0;
       ANEG   <=  1'b    0;
       R_REG  <= 16'h 0000;
       RNZERO <=  1'b    0;

       P_REG  <= 16'h 0000;     // reset-address bits 15:10
       I_REG  <= 16'h 0000;

     end
     else begin

       if (enable & new_D)                  D_REG  <= di_bus;
       if (enable & new_A)     {ANZERO,ANEG,A_REG} <= {|ai_bus,ai_bus[15],ai_bus};
       if (enable &(new_R|calling)) {RNZERO,R_REG} <= {|ri_bus,ri_bus};

       if (enable & new_P)                  P_REG  <= pi_bus;
       if (enable & new_I)                  I_REG  <= ii_bus;

     end
   end // regs

   assign cda = ma_bus;
   assign cdo = D_REG;

endmodule // cpu ===================================================================
